{-# LANGUAGE DerivingVia #-}
{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE DataKinds #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE InstanceSigs #-}
module Language.Java.Classfile.ConstantPool.Entry (Entry(..)) where
import GHC.Generics (Generic, Generically(..))
import Language.Java.Classfile.Extractable (Extractable (extract))
import Language.Java.Classfile.Extractable.WithTag (Word8Tag)
import Language.Java.Classfile.ConstantPool.References (Utf8Reference, ClassReference, NameAndTypeReference, MethodHandleReferenceKind, OpaqueReference, BootstrapMethodIndex)
import Data.Int (Int32, Int64)
import Data.Word (Word16)
import Data.Text (Text)
import Language.Java.Classfile.Extract (Extract, bytes)
import qualified Data.Text.Encoding as Text
import qualified Data.ByteString.Lazy as ByteString
import qualified Data.ByteString as StrictByteString

data Entry 
  = Utf8                (Word8Tag 1  (SizedText Word16))
  | Integer             (Word8Tag 3  Int32)
  | Float               (Word8Tag 4  Float)
  | Long                (Word8Tag 5  Int64)
  | Double              (Word8Tag 6  Double)
  | Class               (Word8Tag 7  Utf8Reference)                                 -- name
  | String              (Word8Tag 8  Utf8Reference)                                 -- value
  | FieldRef            (Word8Tag 9  (ClassReference, NameAndTypeReference))        -- containing class, descriptor
  | MethodRef           (Word8Tag 10 (ClassReference, NameAndTypeReference))        -- containing class, descriptor
  | InterfaceMethodRef  (Word8Tag 11 (ClassReference, NameAndTypeReference))        -- containing class, descriptor
  | NameAndType         (Word8Tag 12 (Utf8Reference, Utf8Reference))                -- name, type
  | MethodHandle        (Word8Tag 15 MethodHandleInfo)
  | MethodType          (Word8Tag 16 Utf8Reference)                                 -- descriptor
  | Dynamic             (Word8Tag 17 (BootstrapMethodIndex, NameAndTypeReference))  -- index into the bootstrapmethod attribute array, NameAndType must refer to a field
  | InvokeDynamic       (Word8Tag 18 (BootstrapMethodIndex, NameAndTypeReference))  -- index into the bootstrapmethod attribute array, NameAndType must refer to a method
  | Module              (Word8Tag 19 Utf8Reference)
  | Package             (Word8Tag 20 Utf8Reference)
  deriving stock (Show, Generic)
  deriving Extractable via Generically Entry

data MethodHandleInfo = MethodHandleInfo MethodHandleReferenceKind OpaqueReference
  deriving stock (Show, Generic)
  deriving Extractable via Generically MethodHandleInfo

newtype SizedText sizeType = SizedText Text
  deriving stock (Show)

instance (Integral sizeType, Extractable sizeType) => Extractable (SizedText sizeType) where
  extract :: Extract (SizedText sizeType)
  extract = do
    size <- extract @sizeType
    slice <- StrictByteString.concat . ByteString.toChunks <$> bytes (fromIntegral size) -- TODO: Replace java zeroes with normal zeroes, normalize 2x3 bytes into 4 byte utf8

    case Text.decodeUtf8' slice of
      Left err -> fail $ show err
      Right t -> pure $ SizedText t

